/*
jfyeNG - a Newtek LightWave scene player
Copyright (C) 1998-2003  Arnaud STORQ a.k.a. No Recess/CONDENSE (norecess@planet-d.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#include <config.h>
/******************************************************************************
 * lwsinterp.c - LWS Scene Interpoler                                         *
 ******************************************************************************
 * Project      : jfyeNG                                                      *
 * Base author  : Ernie Wright 16 Nov 00 (www.newtek.com)                     *
 * Update author: Arnaud Storq (norecess@planet-d.net)                        *
 ******************************************************************************/
#include "platform/os/math.h"
#include "lightwave/lws/lwsinterp.h"

/*---------------------------------------------------------------------------- LOCAL DEFINES */
#define BEH_RESET       0
#define BEH_CONSTANT    1
#define BEH_REPEAT      2
#define BEH_OSCILLATE   3
#define BEH_OFFSET      4
#define BEH_LINEAR      5

/*---------------------------------------------------------------------------- RANGE */
static FLOAT32
range(FLOAT32 v, FLOAT32 lo, FLOAT32 hi, INT32 *i)
{
    FLOAT32 v2 = 0.0f;
    FLOAT32 r = 0.0f;

    r = hi-lo;

    if (r==0.0f)
    {
        if (i!=NULL)
        {
            *i = 0;
        }

        return lo;
    }

    v2 = lo+v-r*(FLOAT32) MathFloor((FLOAT32) v/r);

    if (i!=NULL)
    {
        *i =- (INT32) ((v2-v)/r+(v2>v ? 0.5f : -0.5f));
    }

    return v2;
}

/*---------------------------------------------------------------------------- HERMITE */
static void 
hermite(FLOAT32 t, FLOAT32 *h1, FLOAT32 *h2, FLOAT32 *h3, FLOAT32 *h4)
{
    FLOAT32 t2 = 0.0f;
    FLOAT32 t3 = 0.0f;

    t2 = t*t;
    t3 = t*t2;

    *h2 = 3.0f*t2-t3-t3;
    *h1 = 1.0f-*h2;
    *h4 = t3-t2;
    *h3 = *h4-t2+t;
}

/*---------------------------------------------------------------------------- OUTGOING */
static FLOAT32 
outgoing(Key *key0, Key *key1)
{
    FLOAT32 a = 0.0f;
    FLOAT32 b = 0.0f;
    FLOAT32 d = 0.0f;
    FLOAT32 t = 0.0f;
    FLOAT32 out = 0.0f;

    switch (key0->shape)
    {
    case SHAPE_TCB:
        a = (1.0f-key0->tension)*(1.0f+key0->continuity)*(1.0f+key0->bias);
        b = (1.0f-key0->tension)*(1.0f-key0->continuity)*(1.0f-key0->bias);
        d = key1->value-key0->value;

        if (key0->prev)
        {
            t = (key1->time-key0->time)/(key1->time-key0->prev->time);
            out = t*(a*(key0->value-key0->prev->value)+b*d);
        }
        else
        {
            out = b*d;
        }
        break;

    case SHAPE_LINE:
        d = key1->value-key0->value;
        if (key0->prev)
        {
            t = (key1->time-key0->time)/(key1->time-key0->prev->time);
            out = t*(key0->value-key0->prev->value+d);
        }
        else
        {
            out = d;
        }
        break;

    case SHAPE_BEZI:
    case SHAPE_HERM:
        out = key0->param[1];
        if (key0->prev)
        {
            out *= (key1->time-key0->time)/(key1->time-key0->prev->time);
        }
        break;

    case SHAPE_BEZ2:
        out = key0->param[3]*(key1->time-key0->time);
        if (MathFabs(key0->param[2])>1e-5f)
        {
            out /= key0->param[2];
        }
        else
        {
            out *= 1e5f;
        }
        break;

    case SHAPE_STEP:
    default:
        out = 0.0f;
        break;
    }

    return out;
}

/*---------------------------------------------------------------------------- INCOMING */
static FLOAT32 
incoming(Key *key0, Key *key1)
{
    FLOAT32 a = 0.0f;
    FLOAT32 b = 0.0f;
    FLOAT32 d = 0.0f;
    FLOAT32 t = 0.0f;
    FLOAT32 in = 0.0f;

    switch (key1->shape)
    {
    case SHAPE_LINE:
        d = key1->value-key0->value;
        if (key1->next)
        {
            t = (key1->time-key0->time)/(key1->next->time-key0->time);
            in = t*(key1->next->value-key1->value+d);
        }
        else
        {
            in = d;
        }
        break;

    case SHAPE_TCB:
        a = (1.0f-key1->tension)*(1.0f-key1->continuity)*(1.0f+key1->bias);
        b = (1.0f-key1->tension)*(1.0f+key1->continuity)*(1.0f-key1->bias);
        d = key1->value-key0->value;
        if (key1->next)
        {
            t = (key1->time-key0->time)/(key1->next->time-key0->time);
            in = t*(b*(key1->next->value-key1->value)+a*d);
        }
        else
        {
            in = a*d;
        }
        break;

    case SHAPE_BEZI:
    case SHAPE_HERM:
        in = key1->param[0];
        if (key1->next)
        {
            in *= (key1->time-key0->time)/(key1->next->time-key0->time);
        }
        break;

    case SHAPE_BEZ2:
        in = key0->param[1]*(key1->time-key0->time);
        if (MathFabs(key0->param[0])>1e-5f)
        {
            in /= key0->param[0];
        }
        else
        {
            in *= 1e5f;
        }
        break;

    case SHAPE_STEP:
    default:
        in = 0.0f;
        break;
    }

    return in;
}

/*---------------------------------------------------------------------------- BEZIER */
static FLOAT32 
bezier(FLOAT32 x0, FLOAT32 x1, FLOAT32 x2, FLOAT32 x3, FLOAT32 t)
{
    FLOAT32 a = 0.0f;
    FLOAT32 b = 0.0f;
    FLOAT32 c = 0.0f;
    FLOAT32 t2 = 0.0f;
    FLOAT32 t3 = 0.0f;

    t2 = t*t;
    t3 = t2*t;

    c = 3.0f*(x1-x0);
    b = 3.0f*(x2-x1)-c;
    a = x3-x0-c-b;

    return a*t3+b*t2+c*t+x0;
}

/*---------------------------------------------------------------------------- BEZIER2 TIME */
static FLOAT32 
bez2_time(FLOAT32 x0, FLOAT32 x1, FLOAT32 x2, FLOAT32 x3, FLOAT32 time, FLOAT32 *t0, FLOAT32 *t1)
{
    FLOAT32 v = 0.0f;
    FLOAT32 t = 0.0f;

    t = *t0+(*t1-*t0)* 0.5f;
    v = bezier(x0, x1, x2, x3, t);

    if (MathFabs(time-v)>0.0001f)
    {
        if (v>time)
        {
            *t1 = t;
        }
        else
        {
            *t0 = t;
        }

        return bez2_time(x0, x1, x2, x3, time, t0, t1);
    }

    return t;
}

/*---------------------------------------------------------------------------- BEZ2 */
static FLOAT32 
bez2(Key *key0, Key *key1, FLOAT32 time)
{
    FLOAT32 x = 0.0f;
    FLOAT32 y = 0.0f;
    FLOAT32 t = 0.0f;
    FLOAT32 t0 = 0.0f;
    FLOAT32 t1 = 1.0f;

    if (key0->shape==SHAPE_BEZ2)
    {
        x = key0->time+key0->param[2];
    }
    else
    {
        x = key0->time+(key1->time-key0->time)/3.0f;
    }

    t = bez2_time(key0->time, x, key1->time+key1->param[0], key1->time, time, &t0, &t1);

    if (key0->shape==SHAPE_BEZ2)
    {
        y = key0->value+key0->param[3];
    }
    else
    {
        y = key0->value+outgoing(key0, key1)/3.0f;
    }

    return bezier(key0->value, y, key1->param[1]+key1->value, key1->value, t);
}

/*---------------------------------------------------------------------------- ENVELOPE */
FLOAT32 EnvEval(Env *env, FLOAT32 time)
{
    Key *key0 = NULL;
    Key *key1 = NULL;
    Key *skey = NULL;
    Key *ekey = NULL;

    FLOAT32 t = 0.0f;
    FLOAT32 h1 = 0.0f;
    FLOAT32 h2 = 0.0f;
    FLOAT32 h3 = 0.0f;
    FLOAT32 h4 = 0.0f;
    FLOAT32 in = 0.0f;
    FLOAT32 out = 0.0f;
    FLOAT32 offset = 0.0f;
    INT32 noff = 0;

    if (env->nkeys==0)
    {
        return 0.0f;
    }

    if (env->nkeys==1)
    {
        return env->key->value;
    }

    ekey = env->key;
    skey = ekey;

    while (ekey->next!=NULL)
    {
        ekey = ekey->next;
    }

    if (time<skey->time)
    {
        switch (env->behavior[0])
        {
        case BEH_RESET:
            return 0.0f;

        case BEH_CONSTANT:
            return skey->value;

        case BEH_REPEAT:
            time = range(time, skey->time, ekey->time, NULL);
            break;

        case BEH_OSCILLATE:
            time = range(time, skey->time, ekey->time, &noff);

            if (noff%2)
            {
                time = ekey->time-skey->time-time;
            }

            break;

        case BEH_OFFSET:
            time = range(time, skey->time, ekey->time, &noff);
            offset = noff*(ekey->value-skey->value);
            break;

        case BEH_LINEAR:
            out = outgoing(skey, skey->next)/(skey->next->time-skey->time);
            return out*(time-skey->time)+skey->value;
        }
    }
    else
    {
        if (time>ekey->time)
        {
            switch (env->behavior[1])
            {
            case BEH_RESET:
                return 0.0f;

            case BEH_CONSTANT:
                return ekey->value;

            case BEH_REPEAT:
                time = range(time, skey->time, ekey->time, NULL);
                break;

            case BEH_OSCILLATE:
                time = range(time, skey->time, ekey->time, &noff);
                
                if(noff%2)
                {
                    time = ekey->time-skey->time-time;
                }

                break;

            case BEH_OFFSET:
                time = range(time, skey->time, ekey->time, &noff);
                offset = noff*(ekey->value-skey->value);
                break;

            case BEH_LINEAR:
                in = incoming(ekey->prev, ekey)/(ekey->time-ekey->prev->time);
                return in*(time-ekey->time)+ekey->value;
            }
        }
    }

    key0 = env->key;
    while (time>key0->next->time)
    {
        key0=key0->next;
    }
    key1 = key0->next;

    if (time==key0->time)
    {
        return key0->value+offset;
    }
    else if (time==key1->time)
    {
        return key1->value+offset;
    }

    t = (time-key0->time)/(key1->time-key0->time);

    switch (key1->shape)
    {
    case SHAPE_TCB:
    case SHAPE_BEZI:
    case SHAPE_HERM:
        out = outgoing(key0, key1);
        in = incoming(key0, key1);
        hermite(t, &h1, &h2, &h3, &h4);
        return h1*key0->value+h2*key1->value+h3*out+h4*in+offset;

    case SHAPE_BEZ2:
        return bez2(key0, key1, time)+offset;

    case SHAPE_LINE:
        return key0->value+t*(key1->value-key0->value)+offset;

    case SHAPE_STEP:
        return key0->value+offset;

    default:
        return offset;
    }
}
