/*
jfyeNG - a Newtek LightWave scene player
Copyright (C) 1998-2003  Arnaud STORQ a.k.a. No Recess/CONDENSE (norecess@planet-d.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#include <config.h>
/******************************************************************************
 * matrix.c - Matrix                                                          *
 ******************************************************************************
 * Project      : jfyeNG                                                      *
 * Author       : Arnaud Storq (norecess@planet-d.net)                        *
 ******************************************************************************/
#include "platform/os/ps2/ps2matrix.h"
#include "platform/os/math.h"
#include "platform/renderer/matrix.h"
#include "platform/os/console.h"
#include "system/mem.h"

T_MATRIX matTmp __attribute__((aligned(16)));	

/*---------------------------------------------------------------------------- LOCAL STRUCT */
typedef struct MatrixNode
{
    T_MATRIX data;

    struct MatrixNode *next;
    struct MatrixNode *prev;
} MatrixNode;

/*---------------------------------------------------------------------------- LOCAL VARIABLES */
static MatrixNode *nodeModelView = NULL;
static MatrixNode *nodeProjection = NULL;
static MatrixNode *nodeTexture = NULL;
static MatrixNode *nodeCurrent = NULL;

static INT32 matrixMode = 0;

static T_MATRIX *matrix = NULL;
static T_MATRIX *matrixBase = NULL;

static FLOAT32 radAngle = 0.0f;
static FLOAT32 sinAngle = 0.0f;
static FLOAT32 cosAngle = 0.0f;

/*---------------------------------------------------------------------------- LOCAL FUNCTIONS */
/*---------------------------------------------------------------------------- IDENTITY */
static void
matrixIdentity(void)
{
    memset(matrix, 0, sizeof(T_MATRIX));

    (*matrix)[0][0] = 1.0f;
    (*matrix)[1][1] = 1.0f;
    (*matrix)[2][2] = 1.0f;
    (*matrix)[3][3] = 1.0f;
}

/*---------------------------------------------------------------------------- X ROTATE */
static void
matrixRotateX(void)
{
    matrixIdentity();

    (*matrix)[1][1] = cosAngle;
    (*matrix)[1][2] = sinAngle;
    (*matrix)[2][1] = -sinAngle;
    (*matrix)[2][2] = cosAngle;
}

/*---------------------------------------------------------------------------- Y ROTATE */
static void
matrixRotateY(void)
{
    matrixIdentity();

    (*matrix)[0][0] = cosAngle;
    (*matrix)[0][2] = -sinAngle;
    (*matrix)[2][0] = sinAngle;
    (*matrix)[2][2] = cosAngle;
}

/*---------------------------------------------------------------------------- Z ROTATE */
static void
matrixRotateZ(void)
{
    matrixIdentity();

    (*matrix)[0][0] = cosAngle;
    (*matrix)[0][1] = sinAngle;
    (*matrix)[1][0] = -sinAngle;
    (*matrix)[1][1] = cosAngle;
}

/*---------------------------------------------------------------------------- TRANSLATE */
static void
matrixTranslate(FLOAT32 x, FLOAT32 y, FLOAT32 z)
{
    matrixIdentity();

    (*matrix)[0][3] = x;
    (*matrix)[1][3] = y;
    (*matrix)[2][3] = z;
}

/*---------------------------------------------------------------------------- SCALE */
static void
matrixScale(FLOAT32 x, FLOAT32 y, FLOAT32 z)
{
    matrixIdentity();

    (*matrix)[0][0] = x;
    (*matrix)[1][1] = y;
    (*matrix)[2][2] = z;
}

/*---------------------------------------------------------------------------- FRUSTUM */
static void
matrixFrustum(FLOAT32 left, FLOAT32 right, FLOAT32 bottom, FLOAT32 top, FLOAT32 zNear, FLOAT32 zFar)
{
    matrixIdentity();

	left /= 2.0f;
    right /= 2.0f;
	top /= 2.0f;
	bottom /= 2.0f;
	
	(*matrix)[0][0] = (2.0f*zNear)/(right-left);
    (*matrix)[0][2] = (right+left)/(right-left);
    
    (*matrix)[1][1] = (2.0f*zNear)/(top-bottom);
    (*matrix)[1][2] = (top+bottom)/(top-bottom);
    
    (*matrix)[2][2] = -(zFar+zNear)/(zFar-zNear);
    (*matrix)[2][3] = -(2.0f*zFar*zNear)/(zFar-zNear);

    (*matrix)[3][2] = -1.0f;
}

/*---------------------------------------------------------------------------- ORTHO */
static void
matrixOrtho(FLOAT32 left, FLOAT32 right, FLOAT32 bottom, FLOAT32 top, FLOAT32 zNear, FLOAT32 zFar)
{
    matrixIdentity();

	left /= 2.0f;
    right /= 2.0f;
	top /= 2.0f;
	bottom /= 2.0f;
	
	(*matrix)[0][0] = 2.0f/(right-left);
    (*matrix)[0][3] = (right+left)/(right-left);

    (*matrix)[1][1] = 2.0f/(top-bottom);
    (*matrix)[1][3] = (top+bottom)/(top-bottom);

    (*matrix)[2][2] = -2.0f/(zFar-zNear);
    (*matrix)[2][3] = (zFar+zNear)/(zFar-zNear);

    (*matrix)[3][3] = 1.0f;
}

/*---------------------------------------------------------------------------- INIT */
void
MatrixInit(void)
{
    matrix = (T_MATRIX *) MemAlloc(sizeof(T_MATRIX));
	matrixBase = (T_MATRIX *) MemAlloc(sizeof(T_MATRIX));

	matrixMode = MATRIX_MODELVIEW;
    nodeModelView = (MatrixNode *) MemAlloc(sizeof(MatrixNode));
    nodeCurrent = nodeModelView;
    MatrixLoadIdentity();

    matrixMode = MATRIX_PROJECTION;
    nodeProjection = (MatrixNode *) MemAlloc(sizeof(MatrixNode));
    nodeCurrent = nodeProjection;
    MatrixLoadIdentity();
    
    matrixMode = MATRIX_TEXTURE;
    nodeTexture = (MatrixNode *) MemAlloc(sizeof(MatrixNode));
    nodeCurrent = nodeTexture;
    MatrixLoadIdentity();

    MatrixMode(MATRIX_MODELVIEW);
}

/*---------------------------------------------------------------------------- LOAD IDENTITY */
void
MatrixLoadIdentity(void)
{
    matrixIdentity();

    MatrixLoad(matrix);
}

/*---------------------------------------------------------------------------- ROTATE */
void 
MatrixRotate(FLOAT32 angle, FLOAT32 x, FLOAT32 y, FLOAT32 z)
{
    if (x!=0.0f)
    {
        radAngle = RADIAN(-angle*x);

        sinAngle = (FLOAT32) MathSin(radAngle);
        cosAngle = (FLOAT32) MathCos(radAngle);
    
        matrixRotateX();
    } 
    else if (y!=0.0f)
    {
        radAngle = RADIAN(-angle*y);

        sinAngle = (FLOAT32) MathSin(radAngle);
        cosAngle = (FLOAT32) MathCos(radAngle);
    
        matrixRotateY();
    } 
    else if (z!=0.0f)
    {
        radAngle = RADIAN(-angle*z);

        sinAngle = (FLOAT32) MathSin(radAngle);
        cosAngle = (FLOAT32) MathCos(radAngle);
    
        matrixRotateZ();
    }

    MatrixMult(matrix);
}

/*---------------------------------------------------------------------------- TRANSLATE */
void
MatrixTranslate(FLOAT32 x, FLOAT32 y, FLOAT32 z)
{
    matrixTranslate(x, y, z);

    MatrixMult(matrix);
}

/*---------------------------------------------------------------------------- SCALE */
void
MatrixScale(FLOAT32 x, FLOAT32 y, FLOAT32 z)
{
    matrixScale(x, y, z);

    MatrixMult(matrix);
}

/*---------------------------------------------------------------------------- FRUSTUM */
void
MatrixFrustum(FLOAT32 left, FLOAT32 right, FLOAT32 bottom, FLOAT32 top, FLOAT32 zNear, FLOAT32 zFar)
{
    matrixFrustum(left, right, bottom, top, zNear, zFar);

    MatrixLoad(matrix);
}

/*---------------------------------------------------------------------------- ORTHO */
void
MatrixOrtho(FLOAT32 left, FLOAT32 right, FLOAT32 bottom, FLOAT32 top, FLOAT32 zNear, FLOAT32 zFar)
{
    matrixOrtho(left, right, bottom, top, zNear, zFar);

    MatrixLoad(matrix);
}

/*---------------------------------------------------------------------------- MULTIPLICATION */
void
MatrixMult(T_MATRIX *m)
{
    T_MATRIX *matrixNode = &nodeCurrent->data;
	
	PS2MatrixMult(matrixBase, m, matrixNode);

	MatrixLoad(matrixBase);
}

/*---------------------------------------------------------------------------- VECTOR MULTIPLICATION */
void 
MatrixVectMult(T_VEC *res, T_VEC *in, T_MATRIX *mat)
{
    (*res)[0] = (*in)[0]*(*mat)[0][0] + (*in)[1]*(*mat)[0][1] + (*in)[2]*(*mat)[0][2] + (*mat)[0][3];
    (*res)[1] = (*in)[0]*(*mat)[1][0] + (*in)[1]*(*mat)[1][1] + (*in)[2]*(*mat)[1][2] + (*mat)[1][3];
    (*res)[2] = (*in)[0]*(*mat)[2][0] + (*in)[1]*(*mat)[2][1] + (*in)[2]*(*mat)[2][2] + (*mat)[2][3];
}

void 
MatrixVectMultEnvMap(T_VEC *res, T_VEC *in, T_MATRIX *mat)
{
    (*res)[0] = (*in)[0]*(*mat)[0][0] + (*in)[1]*(*mat)[0][1] + (*in)[2]*(*mat)[0][2];
    (*res)[1] = (*in)[0]*(*mat)[1][0] + (*in)[1]*(*mat)[1][1] + (*in)[2]*(*mat)[1][2];
}

void 
MatrixVectMultWithW(T_VEC *res, T_VEC *in, T_MATRIX *mat)
{
    (*res)[0] = (*in)[0]*(*mat)[0][0] + (*in)[1]*(*mat)[0][1] + (*in)[2]*(*mat)[0][2] + (*mat)[0][3];
    (*res)[1] = (*in)[0]*(*mat)[1][0] + (*in)[1]*(*mat)[1][1] + (*in)[2]*(*mat)[1][2] + (*mat)[1][3];
    (*res)[2] = (*in)[0]*(*mat)[2][0] + (*in)[1]*(*mat)[2][1] + (*in)[2]*(*mat)[2][2] + (*mat)[2][3];
    (*res)[3] = (*in)[0]*(*mat)[3][0] + (*in)[1]*(*mat)[3][1] + (*in)[2]*(*mat)[3][2] + (*mat)[3][3];
}

void 
MatrixVectMultTable(T_VEC *dst, INT32 dstPad, T_MATRIX *mat, T_VEC *src, INT32 srcPad, INT32 count)
{
	INT32 c = 0;
    INT32 l = 0;

	srcPad = 0;
	dstPad = 0;

	for (c=0; c<4; c++)
    {
        for (l=0; l<4; l++)
		{
			matTmp[c][l] = (*mat)[l][c];
		}
	}
    
    PS2MatrixVectMultTable(dst, &matTmp, src, count);
}

void 
MatrixVectMultWithWTable(T_VEC *dst, INT32 dstPad, T_MATRIX *mat, T_VEC *src, INT32 srcPad, INT32 count)
{
	srcPad = 0;
	dstPad = 0;
	
	PS2MatrixVectMultWithWTable(dst, mat, src, count);
}

/*---------------------------------------------------------------------------- GET */
T_MATRIX *
MatrixGet(INT32 mode)
{
    switch(mode)
    {
    case MATRIX_MODELVIEW:
        return &nodeModelView->data;        
        
    case MATRIX_PROJECTION:
        return &nodeProjection->data;

    case MATRIX_TEXTURE:
        return &nodeTexture->data;
    }

    return NULL;
}

/*---------------------------------------------------------------------------- SET */
void 
MatrixLoad(T_MATRIX *m)
{
    memcpy(nodeCurrent->data, m, sizeof(T_MATRIX));
}

/*---------------------------------------------------------------------------- STACK */
void
MatrixPush(void)
{
    MatrixNode *node = NULL;
    
    node = (MatrixNode *) MemAlloc(sizeof(MatrixNode));
    
    switch(matrixMode)
    {
    case MATRIX_MODELVIEW:
        memcpy(node->data, nodeModelView->data, sizeof(T_MATRIX));

        nodeModelView->next = node;
        node->prev = nodeModelView;

        nodeModelView = node;
        break;

    case MATRIX_PROJECTION:
        memcpy(node->data, nodeProjection->data, sizeof(T_MATRIX));

        nodeProjection->next = node;
        node->prev = nodeProjection;

        nodeProjection = node;
        break;

    case MATRIX_TEXTURE:
        memcpy(node->data, nodeTexture->data, sizeof(T_MATRIX));

        nodeTexture->next = node;
        node->prev = nodeTexture;

        nodeTexture = node;
        break;
    }
}

void
MatrixPop(void)
{
    switch(matrixMode)
    {
    case MATRIX_MODELVIEW:
        nodeModelView = nodeModelView->prev;
        
        MemFree(nodeModelView->next);
        nodeModelView->next = NULL;
        
        MatrixLoad(&nodeModelView->data);        
        break;

    case MATRIX_PROJECTION:
        nodeProjection = nodeProjection->prev;
        
        MemFree(nodeProjection->next);
        nodeProjection->next = NULL;
        
        MatrixLoad(&nodeProjection->data);
        break;

    case MATRIX_TEXTURE:
        nodeTexture = nodeTexture->prev;
        
        MemFree(nodeTexture->next);
        nodeTexture->next = NULL;
        
        MatrixLoad(&nodeTexture->data);
        break;
    }
}

/*---------------------------------------------------------------------------- MODE */
void
MatrixMode(INT32 mode)
{
    matrixMode = mode;

    switch(matrixMode)
    {
    case MATRIX_MODELVIEW:
        nodeCurrent = nodeModelView;
        break;

    case MATRIX_PROJECTION:
        nodeCurrent = nodeProjection;
        break;

    case MATRIX_TEXTURE:
        nodeCurrent = nodeTexture;
        break;
    }
}

/*---------------------------------------------------------------------------- DUMP TO CONSOLE */
void 
MatrixDump(T_MATRIX *mat)
{
    INT32 l = 0;

    for (l=0; l<4; l++)
    {
        ConsoleAdd("%d %d %d %d", (INT32)((*mat)[l][0]*100.0f), (INT32)((*mat)[l][1]*100.0f), (INT32)((*mat)[l][2]*100.0f), (INT32)((*mat)[l][3]*100.0f));
    }
    
    ConsoleAdd("");
}