/*
jfyeNG - a Newtek LightWave scene player
Copyright (C) 1998-2003  Arnaud STORQ a.k.a. No Recess/CONDENSE (norecess@planet-d.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#include <config.h>
/******************************************************************************
 * filearc.c - File Archiver                                                  *
 ******************************************************************************
 * Project      : jfyeNG                                                      *
 * Author       : Arnaud Storq (norecess@planet-d.net)                        *
 ******************************************************************************/
#include "platform/os/fileio.h"
#include "platform/os/console.h"
#include "system/filearc.h"
#include "system/filepack.h"
#include "system/mem.h"

/*---------------------------------------------------------------------------- LOCAL DEFINES */
#define ARC_SIGNATURE           "DAT1"
#define DATA_RAW                0
#define DATA_PACKED_HUFFMAN     1

/*---------------------------------------------------------------------------- LOCAL STRUCT */
typedef struct FileArcHeader
{
    INT32 signature;
    INT32 nodeCount;
} FileArcHeader;

typedef struct FileArcNode
{
    INT32 headerSize;
    INT32 compression;
    INT32 fileSize;
    INT32 _unused;
    INT32 fileOffset;
} FileArcNode;

/*---------------------------------------------------------------------------- LOCAL VARIABLES */
static const UINT8 arcSignature[] = {ARC_SIGNATURE};
static UINT8 *arcBuffer = NULL;
static INT32 arcSize = 0;
        
/*---------------------------------------------------------------------------- ARCHIVE CHECK */
static BOOL
isValidArchive(FileArcHeader *header)
{
    if (header->signature==((arcSignature[3]<<24)+(arcSignature[2]<<16)+(arcSignature[1]<<8)+arcSignature[0]))
    {
        return TRUE;
    }

    return FALSE;
}

/*---------------------------------------------------------------------------- PATH PREPARE */
static void 
strPrepare(INT8 *str)
{
    if (str[1]==':')
    {
        strcpy(str, &str[2]);
    }
        
    while (*str!=0)
    {
        if (*str=='\\')
        {
            *str='/';
        }
        else if (*str>='A' && *str<='Z')
        {
            *str += ('a'-'A');
        }

        str++;
    }
}

/*---------------------------------------------------------------------------- INIT */
RESULT 
FileArcInit(INT8 *fileName)
{
    INT32 handle = 0;
    
    ConsoleAdd("Loading archive (%s)", fileName);
	
	handle = FileIOOpen(fileName, FILE_IO_READ_ONLY);
    
    if (handle<=0)
    {
        ConsoleAdd("Cannot open archive file (%s)", fileName);
        
        return ERR;
    }

    arcSize = FileIOGetSize(handle);
        
    arcBuffer = MemAlloc(arcSize);

    if (arcBuffer==NULL)
    {
        FileIOClose(handle);

        return ERR_NOT_ENOUGH_MEM;
    }

    FileIORead(handle, arcBuffer, arcSize);

    FileIOClose(handle);

    return OK;
}

/*---------------------------------------------------------------------------- FREE */
RESULT 
FileArcFree(void)
{
    if (arcBuffer==NULL)
    {
        return ERR;
    }

    SAFE_FREE(arcBuffer);

    return OK;
}

/*---------------------------------------------------------------------------- ARC PRESENT */
BOOL 
FileArcIsPresent(void)
{
    if (arcBuffer==NULL)
    {
        return FALSE;
    }

    return TRUE;
}

/*---------------------------------------------------------------------------- GET FILE */
RESULT 
FileArcGetFile(void **buffer, INT8 *fileName, INT32 *fileSize)
{
    FileArcHeader arcHeader;
    FileArcNode arcNode;
    INT32 offset = 0;
    INT32 nodeCount = 0;
    INT8 nodeFileName[STR_DEFAULT_SIZE] = {""};
    INT8 searchFileName[STR_DEFAULT_SIZE] = {""};
    
    if (arcBuffer==NULL)
    {
        return ERR;
    }

    memcpy(&arcHeader, &arcBuffer[0], sizeof(FileArcHeader));

    if (!isValidArchive(&arcHeader))
    {
        SAFE_FREE(arcBuffer);

        *buffer = NULL;

        ConsoleAdd("Wrong archive signature on archive file (%s)", fileName);
        
        return ERR_ARCHIVE_SIGNATURE;
    }

    strcpy(searchFileName, fileName);
    strPrepare((INT8 *) &searchFileName);

    arcNode.fileSize = 0;
    arcNode.headerSize = 0;
    arcNode.compression = DATA_RAW;
    arcNode.fileOffset = 0;

    offset = 8;
    nodeCount = 0;
    nodeFileName[0] = 0;

    while (strcmp(searchFileName, nodeFileName)!=0 && nodeCount<=arcHeader.nodeCount)
    {
        offset += arcNode.headerSize;    
        memcpy(&arcNode, &arcBuffer[offset], sizeof(FileArcNode));
        strcpy(nodeFileName, (INT8 *) &arcBuffer[offset+sizeof(FileArcNode)]);

        strPrepare((INT8 *) &nodeFileName);
        
        nodeCount++;
    }

    if (nodeCount>arcHeader.nodeCount)
    {
        *buffer = NULL;

        if (strstr(fileName, "prk")!=NULL)
		{
		}
		else if (strstr(fileName, "PRK")!=NULL)
		{
		}
		else
		{
			ConsoleAdd("Missing file (%s) in archive !", fileName);
		}

        return ERR_ARCHIVE_MISSING_FILE;
    }

    switch(arcNode.compression)
    {
    case DATA_RAW:
        *buffer = MemAlloc(arcNode.fileSize);
        memcpy(*buffer, &arcBuffer[arcNode.fileOffset], arcNode.fileSize);
        *fileSize = arcNode.fileSize;
        break;

    case DATA_PACKED_HUFFMAN:
        *fileSize = FileGetUnpackedSize(&arcBuffer[arcNode.fileOffset]);

        *buffer = MemAlloc(*fileSize);
        
        FileUnpack((UINT8 *) &arcBuffer[arcNode.fileOffset], (UINT8 *) *buffer);

        break;
    }
        
    return OK;
}

/*---------------------------------------------------------------------------- GET NB FILES */
RESULT 
FileArcGetNodeCount(INT32 *nodeCount)
{
    FileArcHeader arcHeader;
    
    if (arcBuffer==NULL)
    {
        return ERR;
    }

    memcpy(&arcHeader, &arcBuffer[0], sizeof(FileArcHeader));

    if (!isValidArchive(&arcHeader))
    {
        SAFE_FREE(arcBuffer);    
        *nodeCount=0;

        ConsoleAdd("Wrong archive signature on archive file !");

        return ERR_ARCHIVE_SIGNATURE;
    }

    *nodeCount = arcHeader.nodeCount;

    return OK;
}

/*---------------------------------------------------------------------------- GET FILENAME USING INDEX */
RESULT 
FileArcGetIndexedFilename(INT8 *fileName, const INT32 nodeIndex)
{
    FileArcHeader arcHeader;
    FileArcNode arcNode;
    INT32 offset = 0;
    INT32 i = 0;
    
    if (arcBuffer==NULL)
    {
        return ERR;
    }

    memcpy(&arcHeader, &arcBuffer[0], sizeof(FileArcHeader));

    if (!isValidArchive(&arcHeader))
    {
        SAFE_FREE(arcBuffer);
        
        ConsoleAdd("Wrong archive signature on archive file !");

        return ERR_ARCHIVE_SIGNATURE;
    }

    arcNode.fileSize = 0;
    arcNode.headerSize = 0;
    arcNode.fileOffset = 0;

    offset = 8;
    i = 0;

    do
    {
        memcpy(&arcNode, &arcBuffer[offset], sizeof(FileArcNode));
        strcpy(fileName, (INT8 *) &arcBuffer[offset+sizeof(FileArcNode)]);
        strPrepare((INT8 *) &fileName);

        offset += arcNode.headerSize;
    
        i++;
    } while (i<=nodeIndex);

    return OK;
}