/*
jfyeNG - a Newtek LightWave scene player
Copyright (C) 1998-2003  Arnaud STORQ a.k.a. No Recess/CONDENSE (norecess@planet-d.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#include <config.h>
/******************************************************************************
 * tex.c - Texture                                                            *
 ******************************************************************************
 * Project      : jfyeNG                                                      *
 * Author       : Arnaud Storq (norecess@planet-d.net)                        *
 ******************************************************************************/
#include "system/mem.h"
#include "texture/tex.h"
#include "texture/texscale.h"
#include "texture/texrgb.h"
#include "texture/textga.h"
#include "texture/texjpg.h"
#include "platform/renderer/texhard.h"
#include "platform/os/console.h"

/*---------------------------------------------------------------------------- LOCAL DEFINES */
#define CODEC_MAX_COUNT     8

#define SCROLL_DOWN         "scdn"
#define SCROLL_UP           "scup"
#define SCROLL_DOWN         "scdn"
#define SCROLL_LEFT         "sclt"
#define SCROLL_RIGHT        "scrt"

/*---------------------------------------------------------------------------- LOCAL STRUCTURES */
typedef struct Codec
{
    INT8 ext[3];

    RESULT (*func)(Tex *tex);
} Codec;

/*---------------------------------------------------------------------------- LOCAL VARIABLES */
static Tex *texTree = NULL;
static INT32 texIndex = 0;

static INT32 codecCount = 0;
static Codec codec[CODEC_MAX_COUNT];

static TexSettings texSettings;

/*---------------------------------------------------------------------------- TO UPPER */
static void 
strPrepare(INT8 *str)
{
    while (*str!=0)
    {
        if (*str>='A' && *str<='Z')
        {
            *str += ('a'-'A');
        }

        str++;
    }
}

/*---------------------------------------------------------------------------- INIT */
RESULT 
TexInit(TexSettings *settings)
{
    if (settings == NULL)
    {
		return ERR_INVALID_PTR;
    }
	
	texSettings = *settings;
		
	if (texTree != NULL)
    {
        return ERR;
    }

    texTree = NULL;
    texIndex = 0;

    codecCount = 0;

	if (texSettings.doScalePower2)
	{
		ConsoleAdd("Pow2 automatic scaling enabled");
	}

	TexScaleSetMaxPower2(texSettings.maxPower2);
	
	TexHardInit();

    return OK;
}

/*---------------------------------------------------------------------------- ADD */
RESULT 
TexAdd(Tex *tex)
{
    INT32 i = 0;
    Tex *texNode = NULL;
    
    if (tex==NULL)
    {
        return ERR_INVALID_PTR;
    }

    strPrepare(tex->fileName);
    
    tex->next = NULL;
    tex->prev = NULL;
    tex->index = texIndex;
    tex->width = 0;
    tex->height = 0;
    tex->buffer = NULL;
    tex->parent = NULL;

	tex->doRGBConvert = texSettings.doRGBConvert;
	tex->doScalePower2 = texSettings.doScalePower2;

    texNode = texTree;

    while (texNode!=NULL)
    {
        INT8 tmp[STR_DEFAULT_SIZE];
        
        strcpy(tmp, texNode->fileName);
        strPrepare(tmp);
            
        if (strcmp(tex->fileName, tmp)==0)
        {
            tex->buffer = NULL;
            tex->scrollType = texNode->scrollType;
            tex->width = texNode->width;
            tex->height = texNode->height;
            tex->parent = texNode;
			texNode = NULL;
        }
        else
        {
            texNode = texNode->next;
        }
    }

    if (tex->parent==NULL)
    {
        if (strstr(tex->fileName, SCROLL_UP)!=NULL)
		{
			tex->scrollType = ScrollUp;
		}
		else if (strstr(tex->fileName, SCROLL_DOWN)!=NULL)
		{
			tex->scrollType = ScrollDown;
		}
		else if (strstr(tex->fileName, SCROLL_RIGHT)!=NULL)
		{
			tex->scrollType = ScrollRight;
		}
		else if (strstr(tex->fileName, SCROLL_LEFT)!=NULL)
		{
			tex->scrollType = ScrollLeft;
		}
		else
		{
			tex->scrollType = ScrollUndef;		
		}

        i = 0;
        while (i<codecCount)
        {
            if (strstr(tex->fileName, codec[i].ext)!=NULL)
            {
                if (codec[i].func(tex)!=OK)
                {
                    ConsoleAdd("Error while reading %s !", tex->fileName);

                    return ERR;
                }

                if (tex->doScalePower2)
				{
					TexScalePower2(tex);
				}

				if (tex->doRGBConvert)
				{
					TexRGBConvert(tex);
				}
                
                i = codecCount;
            }
            
            i++;
        }
    }
    
    if (texTree==NULL)
    {
        texTree = tex;
    }
    else
    {
        texNode = texTree;

        while (texNode->next!=NULL)
        {
            texNode = texNode->next;
        }
        
        tex->prev = texNode;
        texNode->next = tex;
    }

    texIndex++;

    return OK;
}

/*---------------------------------------------------------------------------- FREE */
RESULT
TexFree(Tex *tex)
{
    Tex *texNext = NULL;
    Tex *texNode = NULL;

    texNode = texTree;

    while (texNode!=NULL)
    {
        texNext = texNode->next;

        if (texNode==tex)
        {
            SAFE_FREE(texNode->buffer);

            if (texNode->prev!=NULL)
            {
                texNode->prev->next = texNode->next;
            }
            else
            {
                texTree = texNode->next;
            }

            texNode = NULL;
        }
        else
        {
            texNode = texNext;
        }
    }

    return OK;
}

/*---------------------------------------------------------------------------- FREE ALL */
RESULT
TexFreeAll(void)
{
    Tex *texNext = NULL;
    Tex *texNode = NULL;

    texNode = texTree;

    while (texNode!=NULL)
    {
        texNext = texNode->next;

        SAFE_FREE(texNode->buffer);

        texNode = texNext;
    }

    texTree = NULL;
    texIndex = 0;

    return OK;
}

/*---------------------------------------------------------------------------- HARDWARE */
RESULT
TexUploadAllToHard(void)
{
    Tex *tex = NULL;

    tex = texTree;

    while (tex!=NULL)
    {
        if (tex->parent==NULL)
        {
            TexHardAdd(&tex->id, tex->buffer, tex->width, tex->height, tex->depth);
        }
        
        tex = tex->next;
    }

    tex = texTree;

    while (tex!=NULL)
    {
        if (tex->parent!=NULL)
        {
            tex->id = tex->parent->id;
        }
        
        tex = tex->next;
    }

    return OK;
}

RESULT
TexRemoveAllFromHard(void)
{
    Tex *tex = NULL;

    tex = texTree;

    while (tex!=NULL)
    {
        if(tex->parent==NULL)
        {
            TexHardDelete(&tex->id);
        }

        tex = tex->next;
    }

    return OK;
}

/*---------------------------------------------------------------------------- CODEC */
RESULT
TexRegisterCodec(INT8 *ext, RESULT (*func)(Tex *tex))
{
    INT8 str[4];
    
    if (codecCount == CODEC_MAX_COUNT)
    {
        return ERR;
    }

    if (func==NULL || ext==NULL)
    {
        return ERR_INVALID_PTR;
    }
        
    ConsoleAdd("Registering %s codec", ext);
    
    strcpy(str, ext);
    
    strPrepare(str);
        
    strcpy(codec[codecCount].ext, str);

    codec[codecCount].func = func;

    codecCount++;

    return OK;
}
